/*-
 * Copyright (C)2005..2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: multilar.h,v 1.12 2021/06/06 15:43:33 babolo Exp $
 */

#ifndef _MULTILAR_H_
#define _MULTILAR_H_

__BEGIN_DECLS

# ifndef MULAR_COMPAT
#  warning MULAR_COMPAT not defined
# endif

#define MULAR_SWP1    0x000001 /*      HEX                  */
#define MULAR_SWP2    0x000002 /*      HEX                */
#define MULAR_SWP4    0x000004 /*      HEX                    */
#define MULAR_SWP8    0x000008 /*       HEX            */
#define MULAR_SWAP    0x00000F
#define MULAR_DUAD    0x000010 /*                        */
#define MULAR_HEXU    0x000020 /*  unit  HEX                               */
#define MULAR_CHAR    0x000040 /*  unit  char*                             */
#define MULAR_PDMP    0x000100 /*  - ,  todump ,     */
#define MULAR_PFRE    0x000200 /*  - ,  tofree ,     */
#define MULAR_FIXD    0x000400 /*                          */
#define MULAR_ZERO    0x000800 /*                                    */
#define MULAR_TWIX    0x001000 /*                   */
#define MULAR_HALF    0x002000 /*     mular_div                       */
#define MULAR_STRI    0x004000 /*                           */
#define MULAR_UPPE    0x008000 /*                     */
#define MULAR_OBJ     0x010000 /*  ,                            */

#define MULAR_DISPLINE_N 0x0FFFFFFF /*   mular_displine                         */
#define MULAR_DISPLINE_P 0X10000000 /*                         */
#define MULAR_DISPLINE_D 0x20000000 /*  MULAR_DISPLINE_N - ,    */

#define MULAR_NEXT(A) ((A)->top.index)
#define MULAR_SIZE(A) (((A)->lemax - 1) * sizeof(u_int32_t) + sizeof(mular_descriptor))

#ifdef MULAR_INTERNAL
# ifdef INBUG
#  define MULAR_INBUG(A, I) mular_place((A), (I) * INBUG)
# else
#  define MULAR_INBUG(A, I)
# endif
# define ifMULAR_ZERO   if (mular->flags & MULAR_ZERO)

# define SIZE(A)        (mular->size[(A) - 1])                       /* size_t       */

# define NEXT(A)        ((A).link.next)                              /* mular_indir* */
# define LNEXT(I)       ((I) * sizeof(mular_indir))                  /* size_t       */
# define ND(N, I)       (&NEXT(dubl[N])[I])                          /* mular_indir* */
# define NCP(A, B, I)   {memmove((A), (B), LNEXT(I));}
# define NZR(A, I)      {bzero((A), LNEXT(I));}

# define TERM(A)        ((A).link.term)                              /* void*        */
# define LTERM(I)       ((I) * mular->unit)                          /* size_t       */
# define ATERM(A, I)    (TERM(A) + LTERM(I))                         /* void*        */
# define TD(N, I)       (ATERM((dubl[N]), (I)))                      /* void*        */
# define TCP(A, B, I)   {memmove((A), (B), LTERM(I));}
# define TZR(A, I)      {bzero((A), LTERM(I));}

# define ICONT(A, I)    ((A)->index - NEXT(*(A))[I].index)           /* ssize_t      */
# define ICONN(A, I, N) ((A)->index - ((N) = &NEXT(*(A))[I])->index) /* ssize_t      */
#endif

typedef void (*babolo_tofree)(void*);

typedef struct mular_indir {
    union {
        struct mular_indir *next;
        void               *term;
    } link;               /*                 */
    size_t      index;    /*   + 1    */
} mular_indir;

typedef struct {
    mular_indir top;               /*              */
    size_t      unit;              /*                           */
    babolo_tofree tofree;          /*      */
    void  (*todump)(FILE*, void*); /*              */
    FILE       *debugo;            /* Not used                             */
    void       *sema;              /* Not used                             */
    BLIN_flag   flags;             /*                                 */
    u_int16_t   lemax;             /*        */
    u_int16_t   level;             /*            */
    size_t      size[1];           /*                      */
} mular_descriptor;

#ifdef MULAR_INTERNAL
extern BLIN_flag mular_default;
#endif

mular_descriptor *mular_create  __P((BLIN_flag flags, int lemax, size_t unit, const size_t *size));
void              mular_destroy __P((mular_descriptor *mular));
void             *mular_add     __P((mular_descriptor *mular));
void             *mular_insert  __P((mular_descriptor *mular, size_t indx));
void             *mular_getix   __P((mular_descriptor *mular, size_t indx));
void              mular_dump    __P((mular_descriptor *mular, FILE *out));

int   mular_walk
__P((                mular_descriptor *, size_t mindx, mular_indir *idx, int level,      FILE *out
    , void (*beg)   (mular_descriptor *, size_t mindx, mular_indir *idx, int level,      FILE *out)
    , void (*end)   (mular_descriptor *, size_t mindx, mular_indir *idx, int level, int, FILE *out)
    , int  (*before)(mular_descriptor *, size_t mindx, void *elem, size_t length,        FILE *out)
    , int  (*each)  (mular_descriptor *, size_t mindx, void *elem,                       FILE *out)
    , int  (*after) (mular_descriptor *, size_t mindx, void *elem, size_t length,        FILE *out)
   ));

void *mular_find
__P((                 mular_descriptor *, size_t indx, size_t mindx, mular_indir *idx, int level
    , void  (*beg)   (mular_descriptor *, size_t indx, size_t mindx, mular_indir *idx, int level)
    , void  (*end)   (mular_descriptor *, size_t indx, size_t mindx, mular_indir *idx, int level, void *res)
    , void *(*before)(mular_descriptor *, size_t indx, size_t mindx, mular_indir *idx)
    , void  (*after) (mular_descriptor *, size_t indx, size_t mindx, mular_indir *idx, int level)
   ));

void  mular_displine
__P((mular_descriptor *, size_t reli, size_t mindx, mular_indir *idx, int level, FILE *out, size_t flag));

__END_DECLS

#endif /* !_MULTILAR_H_ */
