/*-
 * Copyright (C)2005..2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2005..2021 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: walk.c,v 1.13 2021/06/06 15:43:33 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  4
#define MULAR_COMPAT VMAJOR
#define MULAR_INTERNAL

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "multilar.h"

static const char
mular_sc[] = ": : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : ";

static const u_char
mular_deswap[] =
{ 0x00, 0x01, 0x03, 0x03, 0x07, 0x07, 0x07, 0x07, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f};

void
mular_displine( mular_descriptor *mular
              , size_t       reli
              , size_t       mindx
              , mular_indir *idx
              , int          level
              , FILE        *out
              , size_t       flag
              ) {
    size_t k, l, m, i, d = flag & MULAR_DISPLINE_N;

    if  (flag & MULAR_DISPLINE_D) {
        i = 1;
    } else {
        i = d;
    }
    for (m = 0; m < i; m++) {
        fprintf(out, "%"BLIN_X"(%"BLIN_U")>", BLIN_I((idx[m].link.next)), idx[m].index);
        if  (level == 1) {
            for (k = 0; k < SIZE(level); k++) {
                fprintf( out, "%s %s%s"
                       , (m ? idx[m - 1].index : mindx) + k == idx[m].index ? "$" : ""
                       , (flag & MULAR_DISPLINE_P && k == reli) ? "." : ""
                       , (flag & MULAR_DISPLINE_D && k == d) ? "|" : ""
                       );
                for (l = 0; l < mular->unit; l++) {
                    fprintf(out, "%02X", ((u_char*)ATERM(idx[m], k))[l]);
            }   }
            fprintf( out, "%s%s%s\n"
                   , (m ? idx[m - 1].index : mindx) + k == idx[m].index ? "$" : ""
                   , (flag & MULAR_DISPLINE_P && k == reli) ? "." : ""
                   , (flag & MULAR_DISPLINE_D && k == d) ? "|" : ""
                   );
        } else {
            for (k = 0; k < SIZE(level); k++) {
                fprintf( out, " %s%s%"BLIN_X"(%"BLIN_U")%s"
                       , (flag & MULAR_DISPLINE_P && k == reli) ? "." : ""
                       , (flag & MULAR_DISPLINE_D && k == d) ? "|" : ""
                       , BLIN_I(NEXT(NEXT(idx[m])[k]))
                       , NEXT(idx[m])[k].index
                       , (NEXT(idx[m])[k].index == idx[m].index) ? "$" : ""
                       );
            }
            fprintf(out, "\n");
}   }   }

void *
mular_find( mular_descriptor *mular
          , size_t  indx
          , size_t  mindx
          , mular_indir *idx
          , int     level
          , void  (*beg)   (mular_descriptor *, size_t, size_t, mular_indir *, int level)
          , void  (*end)   (mular_descriptor *, size_t, size_t, mular_indir *, int level, void *res)
          , void *(*before)(mular_descriptor *, size_t, size_t, mular_indir *)
          , void  (*after) (mular_descriptor *, size_t, size_t, mular_indir *, int)
          ) {
    void *res;
    size_t i;

#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    if  (beg) beg(mular, indx, mindx, idx, level);
    if  (indx < mindx || indx > idx->index) {
        ifBLIN_QX1( " index=%"BLIN_U"(%"BLIN_U"..%"BLIN_U") level=%u size=%"BLIN_U
                  , indx, mindx, idx->index, level, SIZE(level)
                  );
        res = NULL;
    } else if (level > mular->level) {
        errno = EINVAL;
        ifBLIN_QX0("level too big: %u > %u", level, mular->level);
        res = NULL;
    } else if (level == 0) {
        res = NULL;
    } else if (level == 1) {
        if  (idx->index - mindx > SIZE(level)) {
            ifBLIN_QX1( " level=%u size=%"BLIN_U" idx->index-mindx=%"BLIN_U"-%"BLIN_U
                      , level, SIZE(level), idx->index, mindx
                      );
            res = NULL;
        } else {
            res =  before ? before(mular, indx, mindx, idx) : ATERM(*idx, indx - mindx);
        }
    } else {
        ssize_t cont;
        mular_indir *t, *f;
        size_t j, midx, relidx;

        for ( i = 0, j = mindx, cont = 0, t = NULL
            ; i < SIZE(level) && ((cont = ICONN(idx, i, t)) >= 0)
            ; i++
        ) {
            if  (t->index > indx) break;
            if  (!cont) break;
            j = t->index;
        }
        relidx = i++;
        midx = j;
        for (; after && i < SIZE(level) && cont && ((cont = ICONN(idx, i, f)) >= 0);) {
            after(mular, indx, j, f, level - 1);
            i++;
            if  (!cont) break;
            j = f->index;
        }
        res = mular_find(mular, indx, midx, t, level - 1, beg, end, before, after);
    }
    if  (end) end(mular, indx, mindx, idx, level, res);
    return(res);
#   undef blin_internal_flags
}

static void
mular_indxb(mular_descriptor *mular, size_t indx, size_t mindx, mular_indir *idx, int level) {
#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    ifBLIN_QX3( "+ flags=%08X indx=%"BLIN_U"[%"BLIN_U"..%"BLIN_U"] %"BLIN_X"(level=%d)"
              , mular->flags, indx, mindx, idx->index, BLIN_I(idx), level
              );
#   undef blin_internal_flags
}

static void
mular_indxe(mular_descriptor *mular, size_t indx, size_t mindx, mular_indir *idx, int level, void *res) {
#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    (void)indx;
    (void)mindx;
    (void)idx;
    (void)level;
    ifBLIN_QX3("- %"BLIN_X, BLIN_I(res));
#   undef blin_internal_flags
}

void *
mular_getix(mular_descriptor *mular, size_t indx) {
    void *res;

#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    ifBLIN_QX2( "+ flags=%08X i=%"BLIN_U, mular->flags, indx);
    if  (indx < mular->top.index) {
        res = mular_find( mular, indx, 0, &(mular->top), mular->level
                        , &mular_indxb, &mular_indxe, NULL, NULL
                        );
    } else {
        res = NULL;
    }
    ifBLIN_QX2("- %"BLIN_X, BLIN_I(res));
    return(res);
#   undef blin_internal_flags
}

int
mular_walk( mular_descriptor *mular
          , size_t mindx
          , mular_indir *idx
          , int    level
          , FILE  *out
          , void (*beg)   (mular_descriptor *, size_t, mular_indir *, int level, FILE *out)
          , void (*end)   (mular_descriptor *, size_t, mular_indir *, int level, int, FILE *out)
          , int  (*before)(mular_descriptor *, size_t, void *, size_t, FILE *out)
          , int  (*each)  (mular_descriptor *, size_t, void *, FILE *out)
          , int  (*after) (mular_descriptor *, size_t, void *, size_t, FILE *out)
          ) {
    size_t i;
    int ex;

#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    if  (beg) beg(mular, mindx, idx, level, out);
    ex = 0;
    if  (level > mular->level) {
        ifBLIN_QX0(" level too big: %u > %u", level, mular->level);
        errno = EINVAL;
        ex = EX_USAGE;
    } else if (level == 0) {
    } else if (level == 1) {
        if  (before) ex = before(mular, mindx, TERM(*idx), idx->index - mindx, out);
        for (i = 0; each && !ex && i + mindx < idx->index; i++) {
            if  (i >= SIZE(level)) {
                ifBLIN_QX0( " %"BLIN_U"    0..%"BLIN_U"(level=%d)"
                          , i, SIZE(level), level
                          );
                ex = EX_USAGE;
                break;
            }
            ex = each(mular, mindx + i, ATERM(*idx, i), out);
        }
        if  (!ex && after) ex = after(mular, mindx, TERM(*idx), idx->index - mindx, out);
    } else {
        ssize_t cont;
        mular_indir *t;
        size_t j;

        for ( i = 0, j = mindx, cont = 0, t = NULL
            ; !ex && i < SIZE(level) && ((cont = ICONN(idx, i, t)) >= 0)
            ; i++
        ) {
            ex = mular_walk(mular, j, t, level - 1, out, beg, end, before, each, after);
            if  (!cont) break;
            j = t->index;
    }   }
    if  (end) end(mular, mindx, idx, level, ex, out);
    return(ex);
#   undef blin_internal_flags
}

static void
mular_dumpb(mular_descriptor *mular, size_t mindx, mular_indir *idx, int level, FILE *out) {
    size_t k = 2 * (32 - mular->level + level);

#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    if  (mular->flags & MULAR_DUAD) {
        fprintf( out, "%s[%"BLIN_U"..%"BLIN_U"]%"BLIN_X"\n"
               , &mular_sc[k], mindx, idx->index, BLIN_I(idx)
               );
    } else {
        fprintf(out, "%s[%"BLIN_U"..%"BLIN_U"]\n", &mular_sc[k], mindx, idx->index);
    }
    ifBLIN_QX3( "+ flags=%08X i=[%"BLIN_U"..%"BLIN_U"] %"BLIN_X"(level=%u size=%"BLIN_U")"
              , mular->flags, mindx, idx->index, BLIN_I(idx), level, SIZE(level)
              );
#   undef blin_internal_flags
}

static int
mular_dumps(mular_descriptor *mular, size_t mindx, void *elem, FILE *out) {
    size_t i, k = 2 * (32 - mular->level + 1);

#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    if  (mular->flags & MULAR_DUAD) {
        fprintf(out, "%s%"BLIN_X"->", &mular_sc[k - 2], BLIN_I(elem));
    } else {
        fprintf(out, "%s[%"BLIN_U"]->", &mular_sc[k - 2], mindx);
    }
    if  (mular->flags & MULAR_HEXU) {
        if  (mular_deswap[mular->flags & MULAR_SWAP] & BLIN_I(elem)) {
            fprintf(out, "\nAddress not alligned\n");
            return(EX_USAGE);
        } else {
            for (i = 0; i < mular->unit; i++) {
                if  ((i & 3) == 3) {
                    fprintf( out, "%02X ", ((u_char*)elem)[i ^ (mular->flags & MULAR_SWAP)]);
                } else {
                    fprintf( out, "%02X", ((u_char*)elem)[i ^ (mular->flags & MULAR_SWAP)]);
    }   }   }   }
    if  (mular->flags & MULAR_CHAR && *(char**)elem) fprintf(out, "%s", *(char**)elem);
    if  (mular->todump) {
        if  (mular->flags & MULAR_PDMP) mular->todump(out, *(void**)elem);
          else mular->todump(out, elem);
    }
    fprintf(out, "\n");
    return(EX_OK);
#   undef blin_internal_flags
}

static void
mular_dumpe(mular_descriptor *mular, size_t mindx, mular_indir *idx, int level, int ex, FILE *out) {
#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    (void)mindx;
    (void)idx;
    (void)level;
    (void)out;
    ifBLIN_QX3("- %d", ex);
#   undef blin_internal_flags
}

void
mular_dump(mular_descriptor *mular, FILE *out) {
    size_t i;

#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    fprintf( out
           , "################### Begin multilar dump ###################\n"
             "flags=%08X unit=%u tofree=%s level=%d(from %d)\n"
           , mular->flags
           , (u_int)mular->unit
           , mular->tofree ? "YES" : "NO"
           , mular->level
           , mular->lemax
           );
    for (i = mular->lemax; i > 0; --i) {
        if  (i > mular->level) {
            fprintf(out, "            ");
        } else {
            fprintf(out, "    %8u", (int)i);
    }   }
    fprintf(out, "\n");
    for (i = mular->lemax; i > 0; --i)
        fprintf( out, "%6d[%04X]"
               , (int)SIZE(i), i == 1 ? (u_int)mular->unit : (u_int)sizeof(mular_indir)
               );
    fprintf(out, "\n");
    mular_walk( mular, 0, &(mular->top), mular->level, out
              , &mular_dumpb, &mular_dumpe, NULL, &mular_dumps, NULL
              );
    fprintf(out, "###################   End multilar dump ###################\n");
#   undef blin_internal_flags
}

static void
mular_freeb(mular_descriptor *mular, size_t mindx, mular_indir *idx, int level, FILE *out) {
#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    (void)out;
    ifBLIN_QX3( "+ flags=%08X i=[%"BLIN_U"..%"BLIN_U"] level=%u size=%"BLIN_U
              , mular->flags, mindx, idx->index, level, SIZE(level)
              );
#   undef blin_internal_flags
}

static int
mular_freea(mular_descriptor *mular, size_t mindx, void *elem, size_t length, FILE *out) {
    (void)mular;
    (void)mindx;
    (void)length;
    (void)out;
    free(elem);
    return(EX_OK);
}

static int
mular_freec(mular_descriptor *mular, size_t mindx, void *elem, FILE *out) {
    (void)mindx;
    (void)out;
    if  (mular->tofree) (*mular->tofree)(mular->flags & MULAR_PFRE ? *(void**)elem : elem);
    return(EX_OK);
}

static void
mular_freee(mular_descriptor *mular, size_t mindx, mular_indir *idx, int level, int ex, FILE *out) {
    (void)mindx;
    (void)idx;
    (void)level;
    (void)out;
#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    ifBLIN_QX3("- %d", ex);
#   undef blin_internal_flags
}

void
mular_destroy(mular_descriptor *mular) {
#   define blin_internal_flags ((mular ? mular->flags : mular_default) & BLIN_MASK)
    if  (!mular) return;
    ifBLIN_QX2( "+ flags=%08X level=%d unit=%"BLIN_U
              , blin_internal_flags, mular->level, mular->unit
              );
    mular_walk( mular
              , 0
              , &(mular->top)
              , mular->level
              , stderr
              , &mular_freeb
              , &mular_freee
              , NULL
              , &mular_freec
              , (mular->flags & MULAR_OBJ) ? NULL : &mular_freea
              );
    ifBLIN_QX2("- flags=%08X", blin_internal_flags);
    free(mular);
#   undef blin_internal_flags
}
